/*
 * (C) 1995-2001 Clemson University and Argonne National Laboratory.
 *
 * See COPYING in top-level directory.
 */

#include <lib.h>
#include <sockio.h>
#include <string.h>
#ifdef HAVE_LINUX_TYPES_H
#include <linux/types.h>
#endif
#ifdef HAVE_LINUX_DIRENT_H
#include <linux/dirent.h>
#else
	/* IRIX */
	#ifdef HAVE_DIRENT_H
	#include <dirent.h>
	#endif
#endif
#ifdef HAVE_SYSCALL_H
#include <syscall.h>
#endif
#ifdef HAVE_SYS_SYSCALL_H
#include <sys/syscall.h>
#endif

#if 0
/* gnu defines are WRONG. */
#include <direntry.h>
#endif

extern fdesc_p pfds[];
extern jlist_p active_p;
extern sockset socks;

static int do_getdents(int fd, struct pvfs_dirent *buf, int size);

/* pvfs_getdents()
 *
 * This call hooks into the getdents syscall.  It returns structures in
 * the format passed back by the kernel.  This is NOT the format that
 * the libc calls pass back, so you shouldn't call this directly.
 *
 * unlike getdents(2), repeated calls to pvfs_getdents will pass pack the 
 * next diretory entries.
 *
 * d_ino and d_off, and d_rclen  are undefined in the returned struct dirent
 * ( i.e. only d_name is useful )
 */
int pvfs_getdents(int fd, struct dirent *userp, unsigned int count)
{
	int requested_dirs= 0, received_dirs = 0, nbytes;
	int i, newoffset=0;
	struct pvfs_dirent *pde;
	fdesc_p pfd_p = pfds[fd];

	if (fd < 0 || fd >= PVFS_NR_OPEN 
		 || (pfds[fd] && pfds[fd]->fs == FS_RESV)) 
	{
		errno = EBADF;
		return(-1);
	}  

	if (!pfd_p || pfd_p->fs == FS_UNIX) {
		if ( (nbytes=syscall(SYS_getdents, fd, userp, count)) == -1 ) {
			return errno;
		}
		if ( nbytes == 0 ) return nbytes;
		/* linux getdent(2) doesn't update the file descriptor, but
		 * pvfs_getdent will update the fd if it's a pvfs directory.
		 * so we'll make pvfs_getdent do the same thing for both types
		 * of files, even if that makes it differ from getdents(2) */
		
		/* the trick is in finding the offset of the last directory*/
		received_dirs = nbytes/sizeof(struct dirent);
		/* now forget the syscall's nbytes and return our own */
		i=0, nbytes=0;
		do {
			/* getdents system call packs in all the dirent 
			 * structures.  have to rely on d_reclen to find 
			 * next structure.  hence the fun pointer games */
			struct dirent *new_de;
			new_de = (struct dirent*)((char *)userp + nbytes);
			newoffset = new_de->d_off;
			nbytes+=new_de->d_reclen;
			++i;
		} while ( i < received_dirs);

		if ( lseek(fd, newoffset, SEEK_SET) == -1) {
			return errno;
		}
		return(nbytes);
	}

	if (pfd_p->fs == FS_PVFS) {
		errno = ENOTDIR;
		return(-1);
	}
	requested_dirs = count / sizeof(struct dirent);
	pde = calloc( 1, requested_dirs * sizeof(struct pvfs_dirent) );
	/* mgr speaks in "struct pvfs_dirent"s, while client code speaks 
	 * in "struct dirent" */
	nbytes = do_getdents(fd, pde, requested_dirs);
	received_dirs = nbytes / sizeof(struct pvfs_dirent);
	for(i=0; i< received_dirs; i++) {
		strcpy( (userp+i)->d_name, (pde+i)->name);
		(userp+i)->d_reclen = sizeof(struct dirent);
	}
	free(pde);

	return received_dirs*sizeof(struct dirent);
}

/* DO_GETDENTS() - handles actually getting the getdents data for a PVFS
 * directory.
 *
 * size is "number of pvfs_dirent things", not bytes
 *
 * Returns -1 on failure, the amount (in bytes) read on success.
 */
static int do_getdents(int fd, struct pvfs_dirent *buf, int size)
{
	int i, s, myeno;
	mreq req;
	mack ack;
	struct sockaddr saddr;
	
	memset(&req, 0, sizeof(req));
	memset(&ack, 0, sizeof(ack));

	if (!pfds[fd]->fn_p) {
		/* we need the directory name in terms the manager can understand */
		errno = EINVAL;
		return(-1);
	}

	req.req.getdents.offset = pfds[fd]->fd.off;
	req.req.getdents.length = size*sizeof(struct pvfs_dirent);
	req.uid = getuid();
	req.gid = getgid();
	req.type = MGR_GETDENTS;
	req.dsize = strlen(pfds[fd]->fn_p);

	saddr = pfds[fd]->fd.meta.mgr;
	if ((s = send_mreq_saddr(&saddr, &req, pfds[fd]->fn_p, &ack)) < 0) {
		myeno = errno;
		PERROR("do_getdents: send_mreq_saddr");
		errno = myeno;
		return(-1);
	}
	if (ack.status) /* error */ {
		errno = ack.eno;
		return(-1);
	}

	if (!ack.dsize) /* hit EOF */ {
		return(0);
	}

	if ((i = brecv(s, buf, ack.dsize)) < 0) {
		myeno = errno;
		PERROR("do_getdents: brecv (getting data)");
		errno = myeno;
		return(-1);
	}
	pfds[fd]->fd.off = ack.ack.getdents.offset;
	return(i);
}

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */ 
