/*
 * (C) 1995-2001 Clemson University and Argonne National Laboratory.
 *
 * See COPYING in top-level directory.
 */


#include "../config.h"
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sockio.h>
#include <pvfs.h>

int parse_args(int argc, char **argv);
int usage(int argc, char **argv);

extern char *optarg;
extern int optind;

static char host[256] = "localhost";
static char fs[256] ="/";
static int port_nr = MGR_REQ_PORT;
static int report_space = 0;

int mgr_ping(char *host, int port_nr);
int iod_ping(char *host, int port_nr);
int iod_freespace(char *host, int port_nr, int index);

int main(int argc, char **argv) {
	int i, s, err, iod_err;
	struct mreq req;
	struct mack ack;
	char *data;
	struct iod_info *info;
	
	parse_args(argc, argv);
	if (optind < argc) {
		usage(argc, argv);
	}

	/*
	 * FIRST DO A SIMPLE PING OF THE MGR
	 */
	err = mgr_ping(host, port_nr);
	if (err < 0) {
		printf("mgr (%s:%d) is down.\n", host, port_nr);
		printf("pvfs file system %s has issues.\n", fs);
		return 1;
	}

	/*
	 * MGR IS UP AND RUNNING; NOW GET IOD INFO
	 */
	s = new_sock();

	req.majik_nr   = MGR_MAJIK_NR;
	req.release_nr = PVFS_RELEASE_NR;
	req.type       = MGR_IOD_INFO;
	req.dsize      = strlen(fs) + 1;

	err = connect_sock(s, host, port_nr);
	if (err < 0) {
		fprintf(stderr, "pvfs-ping: unable to connect to %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto mgr_oops;
	}

	err = bsend(s, &req, sizeof(req));
	if (err < 0) {
		fprintf(stderr, "pvfs-ping: unable to send request to %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto mgr_oops;
	}

	err = bsend(s, fs, req.dsize);
	if (err < 0) {
		fprintf(stderr, "pvfs-ping: unable to send request to %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto mgr_oops;
	}

	err = brecv(s, &ack, sizeof(ack));
	if (err < 0) {
		fprintf(stderr, "pvfs-ping: error receiving acknowledgement from %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto mgr_oops;
	}
	if (ack.majik_nr != MGR_MAJIK_NR) {
		fprintf(stderr, "pvfs-ping: received garbage acknowledgement from %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto mgr_oops;
	}

	if (ack.status != 0) {
		printf("error: mgr (%s:%d) is alive, but %s is not recognized as a file system.\n", 
			host, port_nr, fs);
		printf("pvfs file system %s has issues.\n", fs);
		close(s);
		return 1;
	}

	/* make room for data, receive */
	data = malloc(ack.dsize);
	err = brecv(s, data, ack.dsize);
	if (err < 0) {
		printf("error: mgr (%s:%d) is alive but error occurred receiving iod info for %s.\n", 
			host, port_nr, fs);
		printf("pvfs file system %s has issues.\n", fs);
		close(s);
		return 1;
	}

	printf("mgr (%s:%d) is responding.\n", host, port_nr);

	/*
	 * MANAGER IS OK, NOW WE CHECK ON THE IODS
	 */

	info = (struct iod_info *) data;
	iod_err = 0;
	for (i=0; i < ack.ack.iod_info.nr_iods; i++) {
		static char iod_host[]="xxx.xxx.xxx.xxx\0";
		unsigned char *uc = (unsigned char *)&(info[i].addr.sin_addr);
		int iod_port;

		iod_port = ntohs(info[i].addr.sin_port);
		sprintf(iod_host, "%d.%d.%d.%d", uc[0], uc[1], uc[2], uc[3]);

		err = iod_ping(iod_host, iod_port);
		if (err < 0) {
			printf("iod %d (%s:%d) is down.\n", i, iod_host,
				iod_port);
			iod_err++;
		}
		else {
			printf("iod %d (%s:%d) is responding.\n", i,
				iod_host, iod_port);
			if (report_space) {
				err = iod_freespace(iod_host, iod_port, i);
				if (err < 0) {
					printf("iod %d (%s:%d) failed to report free space.\n", i, iod_host, iod_port);
					iod_err++;
				}
			}
		}
	}

	if (iod_err) {
		printf("pvfs file system %s has issues.\n", fs);
		return 1;
	}
	else {
		printf("pvfs file system %s is fully operational.\n", fs);
		return 0;
	}

mgr_oops:
	close(s);
	printf("mgr (%s:%d) is down.\n", host, port_nr);
	return 1;
}


int parse_args(int argc, char **argv)
{
	int c;

	while ((c = getopt(argc, argv, "h:p:f:s")) != EOF) {
		switch(c) {
			case 's':
				report_space = 1;
				break;
			case 'h':
				strncpy(host, optarg, 255);
				host[255] = 0;
				break;
			case 'p':
				port_nr = atoi(optarg);
				break;
			case 'f':
				strncpy(fs, optarg, 255);
				break;
			case '?':
				usage(argc, argv);
				break;
		}
	}
	return 1;
} /* end of PARSE_ARGS() */

/* mgr_ping(host, port) - does a blocking ping to a mgr
 *
 * Returns 0 on success, -1 on failure.
 */
int mgr_ping(char *host, int port_nr)
{
	int s, err;
	struct mreq req;
	struct mack ack;
	
	req.majik_nr   = MGR_MAJIK_NR;
	req.release_nr = PVFS_RELEASE_NR;
	req.type       = MGR_NOOP;
	req.dsize      = 0;

	s = new_sock();

	err = connect_sock(s, host, port_nr);
	if (err < 0) {
		fprintf(stderr, "pvfs-ping: unable to connect to %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto oops;
	}

	err = bsend(s, &req, sizeof(req));
	if (err < 0) {
		fprintf(stderr, "pvfs-ping: unable to send request to %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto oops;
	}

	err = brecv(s, &ack, sizeof(ack));
	if (err < 0) {
		fprintf(stderr, "pvfs-ping: error receiving acknowledgement from %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto oops;
	}
	if (ack.majik_nr != MGR_MAJIK_NR) {
		fprintf(stderr, "pvfs-ping: received garbage acknowledgement from %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto oops;
	}
	if (ack.release_nr != PVFS_RELEASE_NR) {
		fprintf(stderr, "pvfs-ping: release number from mgr at %s:%d (%d.%d.%d) does not match release number built into this binary (%d.%d.%d).\n",
				  host,
				  port_nr,
				  ack.release_nr / 10000,
				  (ack.release_nr / 100) % 10,
				  ack.release_nr % 10,
				  PVFS_RELEASE_NR / 10000,
				  (PVFS_RELEASE_NR / 100) % 10,
				  PVFS_RELEASE_NR % 10);
		fprintf(stderr, "pvfs-ping: an administrator needs to verify that all PVFS servers and clients are at the same release number.\n");
		fflush(stderr);
		goto oops;
	}
	close(s);
	return 0;

oops:
	close(s);
	return -1;
}

/* iod_freespace(host, port_nr)
 *
 * Returns 0 on success, -1 on failure.
 */
int iod_freespace(char *host, int port_nr, int index)
{
	int s, err;
	struct ireq req;
	struct iack ack;
	
	req.majik_nr   = IOD_MAJIK_NR;
	req.release_nr = PVFS_RELEASE_NR;
	req.type       = IOD_STATFS;
	req.dsize      = 0;

	s = new_sock();
	if (s < 0) {
		goto oops;
	}

	err = connect_sock(s, host, port_nr);
	if (err < 0) {
		goto oops;
	}

	err = bsend(s, &req, sizeof(req));
	if (err < 0) {
		goto oops;
	}

	err = brecv(s, &ack, sizeof(ack));
	if (err < 0) {
		goto oops;
	}

	if (ack.majik_nr != IOD_MAJIK_NR) {
		goto oops;
	}

	if (ack.release_nr != PVFS_RELEASE_NR) {
		goto oops;
	}

	if(ack.status != 0) {
		goto oops;
	}

	printf("iod %d (%s:%d): total space = %Ld Mbytes, free space = %Ld Mbytes\n",
			 index, host, port_nr,
			 ack.ack.statfs.tot_bytes / 1048576,
			 ack.ack.statfs.free_bytes / 1048576);

	close(s);
	return 0;

oops:
	close(s);
	return -1;
}

/* iod_ping(host, port) - does a blocking ping to an iod
 *
 * Returns 0 on success, -1 on failure.
 */
int iod_ping(char *host, int port_nr)
{
	int s, err;
	struct ireq req;
	struct iack ack;
	
	req.majik_nr   = IOD_MAJIK_NR;
	req.release_nr = PVFS_RELEASE_NR;
	req.type       = IOD_NOOP;
	req.dsize      = 0;

	s = new_sock();
	if (s < 0) {
		fprintf(stderr, "pvfs-ping: unable to obtain a socket.\n");
		fflush(stderr);
		goto oops;
	}

	err = connect_sock(s, host, port_nr);
	if (err < 0) {
		fprintf(stderr, "pvfs-ping: unable to connect to iod at %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto oops;
	}

	err = bsend(s, &req, sizeof(req));
	if (err < 0) {
		fprintf(stderr, "pvfs-ping: unable to send request to iod at %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto oops;
	}

	err = brecv(s, &ack, sizeof(ack));
	if (err < 0) {
		fprintf(stderr, "pvfs-ping: error receiving acknowledgement from iod at %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto oops;
	}

	if (ack.majik_nr != IOD_MAJIK_NR) {
		fprintf(stderr, "pvfs-ping: received garbage acknowledgement from iod at %s:%d.\n",
				  host, port_nr);
		fflush(stderr);
		goto oops;
	}

	if (ack.release_nr != PVFS_RELEASE_NR) {
		fprintf(stderr, "pvfs-ping: release number from iod at %s:%d (%d.%d.%d) does not match release number built into this binary (%d.%d.%d).\n",
				  host,
				  port_nr,
				  ack.release_nr / 10000,
				  (ack.release_nr / 100) % 10,
				  ack.release_nr % 10,
				  PVFS_RELEASE_NR / 10000,
				  (PVFS_RELEASE_NR / 100) % 10,
				  PVFS_RELEASE_NR % 10);
		fprintf(stderr, "pvfs-ping: an administrator needs to verify that all PVFS servers and clients are at the same release number.\n");
		fflush(stderr);
		goto oops;
	}

	close(s);
	return 0;

oops:
	close(s);
	return -1;
}

int usage(int argc, char **argv)
{
	fprintf(stderr, "Usage: pvfs-ping [-h host] [-p port] [-f metadata_dir] [-s]\n");
	exit(1);
} /* end of USAGE() */

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */ 
